
import java.util.*;
/**
 * Verwaltet die Fahrzeuge auf einer Fahrbahn.
 * 
 * @author Albert Wiedemann
 * @version V1.0
 */
class Fahrbahn extends Ereignisbehandlung
{
    /** x-Position */
    int x;
    /** y-Position */
    int y;
    /** start-Position */
    int start;
    /** ende-Position */
    int ende;
    /** rand-Position */
    int rand;
    /** Wartezeit zur nächsten Fahrzeugerzeugung */
    int wartezeit;
    
    /** erstes Fahrzeug */
    Fahrzeug fahrzeug1;
    /** zweites Fahrzeug */
    Fahrzeug fahrzeug2;
    /** Darstellungssymbol */
    FahrbahnSymbol darstellung;
    /** zuständige Ampel */
    Ampel ampel;

    /**
     * Erzeugt eine leere Fahrbahn ab dem linken Rand unterhalb der Fenstermitte.
     * @param a für diese Richtung zuständige Ampel
     * @param x x-Koordinate der Fahrbahn
     * @param y y-Koordinate der Fahrbahn
     */
    Fahrbahn(Ampel a, int x, int y)
    {
        this.x = x;
        this.y = y;
        ampel = a;
        fahrzeug1 = null;
        fahrzeug2 = null;
        darstellung = new FahrbahnSymbol();
        darstellung.PositionSetzen(x, y);
        start = darstellung.EintrittsPositionGeben();
        ende = darstellung.AustrittsPositionGeben();
        rand = darstellung.RandAbstandGeben();
        wartezeit = 0;
    }

    /**
     * Reiht ein neues Fahrzeug an der Startposition in die Fahrbahn ein.
     * @param f das neue Fahrzeug
     */
    void FahrzeugEinreihen(Fahrzeug f)
    {
        if (fahrzeug1 == null)
        {
            fahrzeug1 = f;
        }
        else if (fahrzeug2 == null)
        {
            fahrzeug2 = f;
        }
        f.PositionSetzen(x + rand, start);
    }

    /**
     * Ermittelt, ob ein Fahrzeug entfernt werden muss
     * @param f das in Frage kommende Fahrzeug
     */
    boolean IstDraussen(Fahrzeug f)
    {
        return f.YKoordinateGeben() < ende;
    }

    /**
     * Ermittelt, ob das Fahrzeug sich nach dem Vorgänger richten oder
     * die Ampel beachten muss.
     * @param ampelabstand Abstand zur Ampel (Haltelinie)
     * @param vorgängerabstand Abstand zum Vorgängerfahrzeug
     * @param f das aktuelle Fahrzeug
     * @return true, wenn der Vorgänger wichtig ist, false, wenn die Ampel zählt
     */
    boolean VorgängerZählt(int ampelabstand, int vorgängerabstand, Fahrzeug f)
    {
        return (ampelabstand < 0) || (vorgängerabstand < ampelabstand) || ampel.IstGrün() ||
               ampel.IstGelb() && (ampelabstand < f.GeschwindigkeitGeben() * 1.8 * 5) && (f.GeschwindigkeitGeben() > 3.0);
    }

    /**
     * Bewegt alle Fahrzeuge der Fahrbahn für eine Sekunde.
     */
    void FahrzeugeBewegen()
    {
        double vvorgänger;
        int wegmax;
        int ampelAbstand;
        
        //1. Fahrzeug
        if (fahrzeug1 != null)
        {
            wegmax = 100000;
            vvorgänger = 100;
            ampelAbstand = fahrzeug1.YKoordinateGeben() - darstellung.HaltelinienPositionGeben();
            if (VorgängerZählt(ampelAbstand, wegmax, fahrzeug1))
            {
                fahrzeug1.Fahren(wegmax, vvorgänger);
            }
            else
            {
                fahrzeug1.Fahren(ampelAbstand, 0.0);
            }
            //2. Fahrzeug
            if (fahrzeug2 != null)
            {
                vvorgänger = fahrzeug1.GeschwindigkeitGeben();
                wegmax = fahrzeug2.YKoordinateGeben() - fahrzeug1.YKoordinateGeben() - fahrzeug1.LängeGeben() - 2;
                if (vvorgänger > 0)
                {
                    wegmax = wegmax - (int) Math.round((fahrzeug2.GeschwindigkeitGeben() + vvorgänger) / 2.0 * 1.8 * 5);
                }
                ampelAbstand = fahrzeug2.YKoordinateGeben() - darstellung.HaltelinienPositionGeben();
                if (VorgängerZählt(ampelAbstand, wegmax, fahrzeug2))
                {
                    fahrzeug2.Fahren(wegmax, vvorgänger);
                }
                else
                {
                    fahrzeug2.Fahren(ampelAbstand, 0.0);
                }
            }
        }
        //Erstes Fahrzeug draußen?
        if ((fahrzeug1 != null) && IstDraussen(fahrzeug1))
        {
            fahrzeug1.Entfernen();
            fahrzeug1 = fahrzeug2;
            fahrzeug2 = null;
        }
    }
    
    /**
     * Setzt die Position der Fahrbahn.
     * @param x x-Position
     * @param y y-Position
     */
    void PositionSetzen(int x, int y)
    {
        this.x = x;
        this.y = y;
        darstellung.PositionSetzen(x, y);
        rand = darstellung.RandAbstandGeben();
    }
    
    /**
     * Ausgabe der Fahrzeuganzahl
     * @return Anzahl der Fahrzeuge auf der Fahrbahn
     */
    int FahrzeuganzahlGeben()                  
    {
        return fahrzeug1 == null ? 0 : fahrzeug2 == null ? 1 : 2;
    }

    /**
     * Fährt die Fahrzeuge und erzeugt gegebenenfalls ein neues Fahrzeug.
     */
    @Override void TaktImpulsAusführen()
    {
        FahrzeugeBewegen();
        if (wartezeit <= 0)
        {
            FahrzeugEinreihen(new Fahrzeug());
            wartezeit = 4;
        }
        else
        {
            wartezeit = wartezeit - 1;
        }
    }
}
